%   [m,ga,z] = VBIDL(X,y,s2,B,pot,tau,opts)
%
% VBIDL - Variational Bayesian Inference using a Double Loop algorithm
%         minimising the variational criterion 
%      phi(pga) = ln|A| + h(ga) + \min_u R(u,ga), where            (1)
%      A       = X'*X/s2 + B'*diag(1./ga)*B, s=B*u and
%      R(u,ga) = (1/s2) * ||X*u-y||^2 + s'*diag(1./ga)*s - 2*b'*s.
%
% The probabilistic model consists of
%   (i)  Gaussian potentials y = X*u + e, e~N(0,s2*I)   and 
%   (ii) super-Gaussian potentials t_i(s_i), pot(s) = \prod_i t_i(tau_i*s_i)
% leading to a posterior P(u) of the form
%    P(u) = (1/Z) N(u|Xu,s2*I) * pot(s) where s = B*u and a normalisation
%    constant Z = \int  N(u|Xu,s2*I) * pot(s) du.
%
% Our variational inference relaxation uses Gaussian individual lower bounds 
% pot(s) \ge exp( b*s -s^2/ga +h(ga)/2 ) to obtain a joint a lower bound
%   Z \ge exp(h(ga)/2) * \int  N(u|Xu,s2*I) * exp( b*s -s^2/ga ) du = Z_{VB}
% having the form of a Gaussian integral that can be written as
%   Z_{VB} \ge C*exp( -phi(ga)/2 ), C = (2*pi)^(n/2) * (2*pi*s2)^(-m/2).
% Here, the variational criterion phi(ga) is to be minimised w.r.t. the
% variational parameters ga. For log-concave models, this constitutes a convex
% variational optimisation problem (1).
% Our approach to solve the problem uses Fenchel duality to decouple ln|A| by
% upper bounding ln|A| \min_z z'*(1./ga) g*(z); a function that is just a sum
% over individual components of ga. The algorithm consists of an outer loop 
% where the decoupling bound with coefficients z is refit and an inner loop
% where the variational criterion is jointly minimised in both u and ga 
% (wich reduces to a PLS problem, pls/pls*.m).
%
% OUTER LOOP: z \approx diag(B'*inv(A)*B) = var(s)
% INNER LOOP: min_u  (1/s2) * ||X*u-y||^2 + pen(s) where r = sign(s)*sqrt(s^2+z)
%              pen(s) = 2*( tau*b*(r-s) - ln pot(tau*r) ), see pen/penFromPot.m.
%
% INPUT
% =====
%  X   [m,n]  measurement matrix or operator
%  y   [m,1]  measurement vector
%  s2  [1,1]  measurement variance
%  B   [q,n]  matrix or operator
%  pot        potential function handle or function name string from pot/pot*.m
%  tau [q,1]  scale parameters of the potentials
%  opts.      optimisation parameters
%      TODO ...
%      nMVM : maximal number of steps = MVM with A=X'*X+B'*D*B     [default 100]
%
% OUTPUT
% ======
%  m   [n,1]  posterior mean estimate                m = mean(u)
%  ga  [q,1]  optimal value of the variational parameters
%  z   [q,1]  posterior marginal variance estimates  z = var(s)
%
% See also potFunctions.m, plsAlgorithms.m.
%
% (c) by Hannes Nickisch, MPI for Biological Cybernetics, 2010 August 05

function [m,ga,z,varargout] = vbidl(X,y,s2,B,pot,tau,opts)

if nargin<6, error('We need at least 6 input arguments'), end
if nargin==6, opts = 1; end

% set default values
if isfield(opts,'outerZinit')                    % initial value for upper bound
  z0 = opts.outerZinit;
else
  z0 = 0.05;     
end
if isfield(opts,'outerNiter')                  % number of outer loop iterations
  nout = opts.outerNiter;
else
  nout = 10;     
end
if isfield(opts,'outerMVM')       % number of outer loop MVMs or Lanczos vectors
  noutMVM = opts.outerMVM;
else
  noutMVM = 50;     
end
if isfield(opts,'outerExact')  % Do we try and compute exact marginal variances?
  outerExact = opts.outerExact;         % B has to be the identity and X numeric
  exactPoss = isnumeric(B) && numel(B)==1 && isnumeric(X);
  if exactPoss, exactPoss = exactPoss && B==1; end      % Is it possible at all?
  if ~exactPoss, error('Exact marginal variances cannot be computed.'), end
else
  outerExact = 0;     
end
if isfield(opts,'innerMVM')              % number of inner loop MVMs or CG steps
  ninMVM = opts.innerMVM;
else
  ninMVM = 50;     
end
pls_opt.nMVM = ninMVM;
if isfield(opts,'innerIt')              % number of inner Newton loops for plsTN
  pls_opt.nit = opts.innerIt;
elsecell(1,nargout-3)
  pls_opt.nit = 15;     
end
if isfield(opts,'innerPLS')      % method for solving the PLS inner loop problem
  pls = opts.innerPLS;
else
  pls = 'plsTN';     
end

% information parameters
if isfield(opts,'outerOutput')        % flag saying whether some output is shown
  output = opts.outerOutput;
else
  output = false;
end
if isfield(opts,'innerOutput')        % flag saying whether some output is shown
  pls_opt.output = opts.innerOutput;
else
  pls_opt.output = false;
end

n = max(size(X,2),size(B,2)); u = zeros(n,1);
q = numel(B*u);
pi = 1;
z  = z0*ones(q,1);

if output, tic, end
for i = 1:nout
  if output, fprintf('BEGIN loop %d/%d\n',i,nout), end
  pi_old = pi;

  % INNER loop: PLS, no ga dependence since implicitely minimised
  u = feval(pls,u,X,y,B,pls_opt,s2,'penFromPot',pot,tau,z);
  
  % INNER loop: optimal ga = 1./pi
  s = B*u; 
  sign_s = 2*(s>=0)-1;                              % strict sign mapping 0 to 1
  r = sign_s.*sqrt(s.*s + z + eps);      % signed stabilised z-smoothed variable
  [lp,dlp,ignore,b] = cols( pot(tau.*r) ); dlp = tau.*dlp; b = tau.*b;
  pi = abs( (b-dlp)./r );

  % OUTER loop: marginal variances
  if i<nout    
    if outerExact                    % exact computation by the Woodbury formula
      z = diaginv_woodbury(X,1/s2,pi);
      for j=1:nargout-3, varargout{3+j} = []; end
    else                                    % approximate computation by Lanczos
      [z,varargout{1:nargout-3}] = diaginv_lanczos(X,1/s2,B,pi,noutMVM,output);
    end
  end

  dpi = norm(pi_old-pi)/norm(pi); cvg = dpi<sqrt(eps);
  if cvg || i==nout % converged or last
    if output, fprintf('  <- algorithm converged(%d/%d)\n',i,nout), end
    if output, fprintf('Compute marginal mean\n'), end
    if outerExact                                                 % compute mean
      m = linsolve_woodbury(X,1/s2,pi,X'*y/s2+B'*b);
    else
      m = linsolve_lcg(X,1/s2,B,pi,[X']*y/s2+[B']*b,ninMVM,u);
    end
    ga = 1./pi;
    break
  end 
  if output
    fprintf('END loop %d/%d: dpi=%1.4e after %1.1fs\n',i,nout,dpi,toc)
  end
  ga = 1./pi;
end

% split a matrix into its columns
function varargout = cols(A)
  ncol = size(A,2); varargout = cell(1,nargout);
  for n=1:nargout, if n>ncol, varargout{n}=[]; else varargout{n}=A(:,n); end,end