function A = cnv2mat(f, sx, shape)
%CNV2MAT creates a matrix object for 2d convolutions.
%
% Inputs:
%   f        the left-hand-side of the convolution
%   sx       the size of the right-hand-side of the convolution
%   shape    one of 'circ', 'same' or 'valid'
%
% Example of usage:
%   F = cnv2mat(f, sx, 'same');
%   y = F*x;    % same as y = imfilter(x, f, 'same');
%
% Michael Hirsch * 20/07/2010


  sf = size(f);        % size of the psf, not of the matrix

  if any(sf < sx) && any(sf > sx)
    error('[%s.m] size missmatch', mfilename);
  end

  sF = max(sx, sf);    % the bigger one
  sw = min(sx, sf);    % the smaller one

  switch shape
    case 'circ'
      sy = sF;
    case 'valid'
      sy = sF - sw + 1;
    case 'same'
      sy = sF;
      sF = sx + sf - 1;
    otherwise
      error('[cnv2mat.m] Unknown shape. Choose between ''circ'', ''same'' and ''valid''');
  end

  % precompute fft already
  F = fftn(f, sF);

  % calculate its size
  sz = [prod(sy), prod(sx)];

  % store all stuff in the structure
  A.shape = shape;    % shape of the convolution
  A.f     = f;        % PSF
  A.F     = F;        % precomputed FFT
  A.sx    = sx;       % input dimension
  A.sy    = sy;       % output dimension
  A.sf    = sf;       % size of the PSF
  A.sF    = sF;     % size of the FFT
  A.sw    = sw;       % size of the PSF

  A.sz = sz; A.ctransp = 0; A = class(A,mfilename);                  % [generic]